<?php
namespace EasyCredit;

require_once 'autoload.php';

use EasyCredit\Config       as Config;
use EasyCredit\Process\Exception\InvalidTransitionException;
use EasyCredit\Process\Process;
use EasyCredit\Log\Handler\NullHandler;
use EasyCredit\Log\Handler\FileHandler;
use EasyCredit\Log\Logger;
use EasyCredit\Log\LogLevel as LogLevel;
use EasyCredit\Api\ApiClient;
use EasyCredit\Api\DataMapper;
use EasyCredit\Http\Request;
use EasyCredit\Http\Adapter\Curl;
use EasyCredit\Transfer\Customer;
use EasyCredit\Transfer\RiskRelatedInfo;
use EasyCredit\Transfer\ProcessData;
use EasyCredit\Transfer\ProcessInitialize;
use EasyCredit\Transfer\TechnicalShopParams;
use EasyCredit\Transfer\DeliveryAddress;
use EasyCredit\Transfer\Contact;
use EasyCredit\Transfer\BillingAddress;
use EasyCredit\Transfer\CartInfoCollection;
use EasyCredit\Transfer\CallbackUrls;
use EasyCredit\Transfer\EmploymentData;
use EasyCredit\Transfer\BankData;
use EasyCredit\Transfer\Agreement;
use EasyCredit\Client\Result\EasyCreditClientInitializeResult;
use EasyCredit\Client\Result\EasyCreditClientBaseResult;
use EasyCredit\Client\Result\EasyCreditClientDecideResult;
use EasyCredit\Client\Result\EasyCreditClientCommonProcessDataResult;
use EasyCredit\Client\Result\EasyCreditClientFinancingDetailsResult;
use EasyCredit\Client\Result\EasyCreditClientDecisionResult;
use EasyCredit\Client\Result\EasyCreditClientLegislativeTextResult;
use EasyCredit\Client\Result\EasyCreditClientResidualAmountLimitResult;
use EasyCredit\Client\Result\EasyCreditClientModelCalculationResult;
use EasyCredit\Client\Result\EasyCreditClientBestModelCalculationResult;
use EasyCredit\Process\Status;

/**
 * Class EasyCreditClient
 * 
 * @package EasyCredit
 */
class EasyCreditClient
{
    /**
     * @var string
     */
    protected $webShopId;

    /**
     * @var string
     */
    protected $webShopToken;

    /**
     * 
     * @var Process
     */
    protected $process;
    
    /**
     * @var \EasyCredit\Api\ApiClient
     */
    protected $apiClient;
    
    /**
     * @var string
     */
    protected $integrationType;
    
    /**
     * EasyCreditClient constructor.
     * 
     * @param string $webShopId
     * @param string $webShopToken
     * @param string $integrationType 
     * @param string $logFilename full path including filename
     * @param string $logLevel error|debug
     */
    public function __construct(
        $webShopId,
        $webShopToken,
        $integrationType = ProcessInitialize::INTEGRATION_TYPE_PAYMENT_PAGE,
        $logFilename = null,
        $logLevel = LogLevel::ERROR
        )
    {
        $this->webShopId       = $webShopId;
        $this->webShopToken    = $webShopToken;
        $this->integrationType = $integrationType;
        
        if ($logFilename == null) {
            $logger = new Logger(new NullHandler(), $logLevel);
        } else {
            $logger = new Logger(new FileHandler($logFilename), $logLevel);
        }
        
        $requestAdapter     = new Curl();
        $requestAdapter->setLogger($logger);

        $request = new Request(
            Config::EASYCREDIT_API_HOSTNAME, Config::EASYCREDIT_API_PORT, $requestAdapter
        );
        $dataMapper = new DataMapper();

        $this->apiClient = new ApiClient(
            $this->webShopId,
            $this->webShopToken,
            $request,
            $dataMapper
        );
        
        try {
            $this->process = Process::createInstance($this->apiClient);
        } catch (Exception $e) {
            $this->process = Process::getInstance();
        }
    }
    

    /**
     * Initialises a new transaction
     * 
     * @param Customer            $customer
     * @param BillingAddress      $billingAddress
     * @param DeliveryAddress     $deliveryAddress
     * @param float               $orderTotal
     * @param RiskRelatedInfo     $riskRelatedInfo
     * @param CartInfoCollection  $cartInfos
     * @param CallbackUrls        $callbackUrls
     * @param int                 $term
     * @param TechnicalShopParams $technicalShopParams
     * @param string              $processIdentifierShop
     * 
     * @return EasyCreditClientInitializeResult
     * 
     * @throws InvalidTransitionException
     */
    public function initialize(
        Customer $customer,
        BillingAddress $billingAddress,
        DeliveryAddress $deliveryAddress,
        $orderTotal,
        RiskRelatedInfo $riskRelatedInfo,
        CartInfoCollection $cartInfos,
        CallbackUrls $callbackUrls,
        $term,
        TechnicalShopParams $technicalShopParams = null,
        $processIdentifierShop = null
        )
    {
        $processData = $this->getProcess()->getProcessData();
        $processData->setCustomer($customer);
        $processData->setBillingAddress($billingAddress);
        $processData->setDeliveryAddress($deliveryAddress);
        $processData->setOrderTotal($orderTotal);
        $processData->setRiskInfo($riskRelatedInfo);
        $processData->setProducts($cartInfos);
        $processData->setCallbackUrls($callbackUrls);
        $processData->setTerm($term);

        $processInitializeResponse = $this->process->initialize($this->integrationType, $technicalShopParams, $processIdentifierShop);
        
        if (!$processInitializeResponse) {
            $httpCode          = null;
            $error             = true;
            $processIdentifier = null;
        } else {
            $httpCode          = $processInitializeResponse->getHttpStatusCode();
            $error             = $processInitializeResponse->getHttpStatusCode() == 201 ? false : true;
            $processIdentifier = $processInitializeResponse->getTbProcessIdentifier();
        }
        
        $messages = $this->process->getMessages();
        
        if (!empty($this->getProcess()->getProcessData()->getMessages())) {
            array_push($messages, $this->getProcess()->getProcessData()->getMessages());
        }
        
        return new EasyCreditClientInitializeResult(
            $httpCode,
            $messages,
            $error,
            $processIdentifier
            );
    }
    
    /**
     * updates specific (optional) customer values
     * 
     * @param int            $term
     * @param EmploymentData $employmentData
     * @param Contact        $contact
     * @param Agreement      $agreement
     * @param BankData       $bankData
     * @param \DateTime      $birthDate
     * 
     * @return EasyCreditClientBaseResult
     * 
     * @throws InvalidTransitionException
     */
    public function update(
        $term = null,
        EmploymentData $employmentData = null,
        Contact $contact = null,
        Agreement $agreement = null,
        BankData $bankData = null,
        \DateTime $birthDate = null
        )
    {
        $processData = $this->getProcess()->getProcessData();
        
        if (null != $term) {
            $processData->setTerm($term);
        }
        
        if (null != $employmentData) {
            $processData->getCustomer()->setEmploymentData($employmentData);
        }
        
        if (null != $contact) {
            $processData->getCustomer()->setContact($contact);
        }
        
        if (null != $contact) {
            $processData->getCustomer()->setContact($contact);
        }

        if (null != $agreement) {
            $processData->setAgreement($agreement);
        }
        
        if (null != $bankData) {
            $processData->setBankData($bankData);
        }
        if (null != $birthDate) {
            $processData->getCustomer()->getPersonData()->setBirthDate($birthDate);
        }
        
        $processUpdatedResponse = $this->process->update($this->integrationType);
        
        if (!$processUpdatedResponse) {
            $httpCode = null;
            $error    = true;
            $updated  = null;
        } else {
            $httpCode = $processUpdatedResponse->getHttpStatusCode();
            $error    = $processUpdatedResponse->getHttpStatusCode() == 200 ? false : true;
            $updated  = $processUpdatedResponse->getHttpStatusCode() == 200 ? true : false;
        }
        $messages = $this->process->getMessages();
        
        return new EasyCreditClientBaseResult(
            $httpCode,
            $messages,
            $error,
            $updated
            );
    }
    
    /**
     * inform easycredit to decide. Result can be "ROT" for denied, "GRUEN" for approved or false for possible error
     * 
     * @return EasyCreditClientDecideResult
     * 
     * @throws InvalidTransitionException
     */
    public function decide()
    {
        $processDecideResponse = $this->process->decide($this->integrationType);
        
        if (!$processDecideResponse) {
            $httpCode = null;
            $error    = true;
            $decision = null;
        } else {
            $httpCode = $processDecideResponse->getHttpStatusCode();
            $error    = $processDecideResponse->getHttpStatusCode() == 200 ? false : true;
            $decision = $processDecideResponse->getResult();
        }
        $messages = $this->process->getMessages();
        
        return new EasyCreditClientDecideResult(
            $httpCode,
            $messages,
            $error,
            $decision
            );
    }
    
    /**
     * verifies a given mTAN; only for integration type service integration
     * 
     * @param string $mTan
     * 
     * @return EasyCreditClientBaseResult
     * 
     * @throws InvalidTransitionException
     */
    public function verifyMTan($mTan)
    {
        $processData = $this->getProcess()->getProcessData();
        $processData->setMTan($mTan);
     
        $processVerifyMtanResponse = $this->process->verifyMtan($this->integrationType);
    
        if (!$processVerifyMtanResponse) {
            $httpCode  = null;
            $error     = true;
            $processed = null;
        } else {
            $httpCode  = $processVerifyMtanResponse->getHttpStatusCode();
            $error     = $processVerifyMtanResponse->getHttpStatusCode() == 200 ? false : true;
            $processed = $processVerifyMtanResponse->getHttpStatusCode() == 200 ? true : false;
        }
        $messages = $this->process->getMessages();
        
        return new EasyCreditClientBaseResult(
            $httpCode,
            $messages,
            $error,
            $processed
            );
    }

    /**
     * confirms (finalises) an approved transaction
     * @param string $processIdentifierShop
     * 
     * @return EasyCreditClientBaseResult
     * 
     * @throws InvalidTransitionException
     */
    public function confirm($processIdentifierShop = null)
    {
        $processConfirmResponse = $this->process->agree($this->integrationType, $processIdentifierShop);
        
        if (!$processConfirmResponse) {
            $httpCode  = null;
            $error     = true;
            $processed = null;
        } else {
            $httpCode  = $processConfirmResponse->getHttpStatusCode();
            $error     = $processConfirmResponse->getHttpStatusCode() == 200 ? false : true;
            $processed = $processConfirmResponse->getHttpStatusCode() == 200 ? true : false;
        }
        $messages = $this->process->getMessages();
        
        return new EasyCreditClientBaseResult(
            $httpCode,
            $messages,
            $error,
            $processed
            );
    }
    
    /**
     * requests common process data for a specific process identifier; 
     * 
     * @param string $tbProcessIdentifier
     * 
     * @return EasyCreditClientCommonProcessDataResult
     */
    public function getCommonProcessData($tbProcessIdentifier = null) {
        
        $commonProcessDataResponse = $this->process->getCommonProcessData($tbProcessIdentifier);
        
        if (!$commonProcessDataResponse) {
            $httpCode          = null;
            $error             = true;
            $commonProcessData = null;
        } else {
            $httpCode          = $commonProcessDataResponse->getHttpStatusCode();
            $error             = $commonProcessDataResponse->getHttpStatusCode() == 200 ? false : true;
            $commonProcessData = $commonProcessDataResponse->getCommonProcessData();
        }
        $messages = $this->process->getMessages();
        
        return new EasyCreditClientCommonProcessDataResult(
            $httpCode,
            $messages,
            $error,
            $commonProcessData
            );
    }
    
    /**
     * resends a mTAN to the customer for a specific process identifier; if no identifier provided, it will use the current process identifier
     * 
     * @param string $tbProcessIdentifier
     * 
     * @return EasyCreditClientBaseResult
     */
    public function resendMTan($tbProcessIdentifier = null)
    {
        $processResendMtanResponse = $this->process->resendMtan($tbProcessIdentifier);
    
        if (!$processResendMtanResponse) {
            $httpCode  = null;
            $error     = true;
            $processed = null;
        } else {
            $httpCode  = $processResendMtanResponse->getHttpStatusCode();
            $error     = $processResendMtanResponse->getHttpStatusCode() == 200 ? false : true;
            $processed = $processResendMtanResponse->getHttpStatusCode() == 200 ? true : false;
        }
        $messages = $this->process->getMessages();
        
        return new EasyCreditClientBaseResult(
            $httpCode,
            $messages,
            $error,
            $processed
            );
    }
    
    /**
     * requests financing details for a specific process identifier; 
     * 
     * @param string $tbProcessIdentifier
     * 
     * @return EasyCreditClientFinancingDetailsResult
     */
    public function getFinancingDetails($tbProcessIdentifier = null)
    {
        $financingDetailsResponse = $this->process->getFinancingDetails($tbProcessIdentifier);
        
        if (!$financingDetailsResponse) {
            $httpCode        = null;
            $error           = true;
            $installmentPlan = null;
            $repaymentText   = null;
        } else {
            $httpCode        = $financingDetailsResponse->getHttpStatusCode();
            $error           = $financingDetailsResponse->getHttpStatusCode() == 200 ? false : true;
            $installmentPlan = $financingDetailsResponse->getInstallmentPlan();
            $repaymentText   = $financingDetailsResponse->getRepaymentPlanText();
        }
        $messages = $this->process->getMessages();
        
        return new EasyCreditClientFinancingDetailsResult(
            $httpCode,
            $messages,
            $error,
            $installmentPlan,
            $repaymentText
            );
    }
    
    /**
     * requests decision for a specific process identifier; if identifier not provided it will use current process identifier
     * 
     * @param string $tbProcessIdentifier
     *
     * @return EasyCreditClientDecisionResult
     */
    public function getDecision($tbProcessIdentifier = null)
    {
        $processDecisionResponse = $this->process->getDecision($tbProcessIdentifier);
        
        if (!$processDecisionResponse) {
            $httpCode = null;
            $error    = true;
            $decision = null;
        } else {
            $httpCode = $processDecisionResponse->getHttpStatusCode();
            $error    = $processDecisionResponse->getHttpStatusCode() == 200 ? false : true;
            $decision = $processDecisionResponse->getDecision();
        }
        $messages = $this->process->getMessages();
        
        return new EasyCreditClientDecisionResult(
            $httpCode,
            $messages,
            $error,
            $decision
            );
    }
    
    /**
     * 
     * @param float             $orderTotal
     * @param BillingAddress    $billingAddress
     * @param DeliveryAddress   $deliveryAddress
     * 
     * @return boolean
     */
    public function isPaymentMethodAvailable($orderTotal, $billingAddress = null, $deliveryAddress = null) {
    
        $isAvailable = false;
        if (Config::isValidOrderAmount($orderTotal)) {
            $isAvailable = true;
            
            if (null != $billingAddress && !Config::isValidCountryIso2Code($billingAddress->getCountryCode())) {
                $isAvailable = false;
            }
            
            if (null != $deliveryAddress && !Config::isValidCountryIso2Code($deliveryAddress->getCountryCode())) {
                $isAvailable = false;
            }
            
            if (null != $deliveryAddress &&
                null != $deliveryAddress &&
                $this->isAlternativeDeliveryAddress($billingAddress, $deliveryAddress)) {
                $isAvailable = false;
            }
        }
        
        return $isAvailable;
    }
    
    /**
     * returns legislative text for given shop id
     * 
     * @return EasyCreditClientLegislativeTextResult
     */
    public function getLegislativeText()
    {
        $processLegislativeTextResponse = $this->process->getLegislativeText();
        
        if (!$processLegislativeTextResponse) {
            $httpCode        = null;
            $error           = true;
            $legislativeText = null;
        } else {
            $httpCode        = $processLegislativeTextResponse->getHttpStatusCode();
            $error           = $processLegislativeTextResponse->getHttpStatusCode() == 200 ? false : true;
            $legislativeText = $processLegislativeTextResponse;
        }
        $messages = $this->process->getMessages();
        
        return new EasyCreditClientLegislativeTextResult(
            $httpCode,
            $messages,
            $error,
            $legislativeText
            );
    }
    
    /**
     * returns the residual amount purchase limit for given shop id
     *
     * @return EasyCreditClientResidualAmountLimitResult
     */
    public function getResidualAmountPurchaseLimit()
    {
        $processResidualAmountLimitResponse = $this->apiClient->getResidualAmountPurchaseLimit();
        
        if (!$processResidualAmountLimitResponse) {
            $httpCode       = null;
            $error          = true;
            $residualAmount = null;
        } else {
            $httpCode       = $processResidualAmountLimitResponse->getHttpStatusCode();
            $error          = $processResidualAmountLimitResponse->getHttpStatusCode() == 200 ? false : true;
            $residualAmount = $processResidualAmountLimitResponse->getResidualAmount();
        }
        $messages = $this->process->getMessages();
        
        return new EasyCreditClientResidualAmountLimitResult(
            $httpCode,
            $messages,
            $error,
            $residualAmount
            );
    }
    
    /**
     * @param float $amount
     * 
     * @return EasyCreditClientModelCalculationResult
     */
    public function getModelCalculation($amount)
    {
        $modelCalculcationResponse = $this->process->getModelCalculation($amount);
        
        if (!$modelCalculcationResponse) {
            $httpCode         = null;
            $error            = true;
            $modelCalculation = null;
        } else {
            $httpCode         = $modelCalculcationResponse->getHttpStatusCode();
            $error            = $modelCalculcationResponse->getHttpStatusCode() == 200 ? false : true;
            $modelCalculation = $modelCalculcationResponse;
        }
        $messages = $this->process->getMessages();
        
        return new EasyCreditClientModelCalculationResult(
            $httpCode,
            $messages,
            $error,
            $modelCalculation
            );
    }
    
    /**
     * @param float $amount
     * 
     * @return EasyCreditClientBestModelCalculationResult
     */
    public function getBestModelCalculation($amount)
    {
        $bestModelCalculcationResponse = $this->apiClient->getBest($amount);
    
        if (!$bestModelCalculcationResponse) {
            $httpCode             = null;
            $error                = true;
            $bestModelCalculation = null;
        } else {
            $httpCode             = $bestModelCalculcationResponse->getHttpStatusCode();
            $error                = $bestModelCalculcationResponse->getHttpStatusCode() == 200 ? false : true;
            $bestModelCalculation = $bestModelCalculcationResponse;
        }
        $messages = $this->process->getMessages();
        
        return new EasyCreditClientBestModelCalculationResult(
            $httpCode,
            $messages,
            $error,
            $bestModelCalculation
            );
    }
    
    /**
     * destroys recent process
     * 
     * @return boolean
     */
    public function destroyProcess()
    {
        try {
            $this->process->destroy();
            return true;
        } catch (Exception $e) {
            return false;
        }
    }
    
    /**
     *
     * @return Process
     */
    protected function getProcess()
    {
        return $this->process;
    }

    /**
     *
     * @param Process $process            
     */
    protected function setProcess(Process $process)
    {
        $this->process = $process;
    }
    
    /**
     * 
     * @param BillingAddress  $billingAddress
     * @param DeliveryAddress $deliveryAddress
     * 
     * @return boolean
     */
    protected function isAlternativeDeliveryAddress($billingAddress, $deliveryAddress)
    {
        if (null == $billingAddress || null == $deliveryAddress) {
            return true;
        }
        
        if (
            $billingAddress->getStreet() != $deliveryAddress->getStreet() ||
            $billingAddress->getAddressAdditional() != $deliveryAddress->getAddressAdditional() ||
            $billingAddress->getZip() != $deliveryAddress->getZip() ||
            $billingAddress->getCity() != $deliveryAddress->getCity() ||
            $billingAddress->getCountryCode() != $deliveryAddress->getCountryCode()
            ) {
                return true;
        }
        
        return false;
    }
}
